// //////////////////////////////////////////////////////////////////! 
// ------------------------------------------------------------------ 
// COPYRIGHT (c) 2004-2016 <VirtualSim> 
// <legal statement> 
// ------------------------------------------------------------------  
// 
// Software Unit Name: vsTASKER  
// 
// ------------------------------------------------------------------  
// All rights reserved. 
// http://www.virtualsim.com 
// //////////////////////////////////////////////////////////////////! 

#ifndef VT_WINGDYN_HIGH_H
#define VT_WINGDYN_HIGH_H

// [!VT]  DEFINITION 
// ------------------ 
// Include directives 
// ------------------ 

#include <vt_components.h> 
#include <cls_types.h> 

#include <hi_alt_ctrl.h>
#include <hi_spd_ctrl.h>
#include <hi_roll_ctrl.h>
#include <hi_pitch_ctrl.h>
#include <hi_hdg_ctrl.h>

#include <basic_dyn.h>
// -------------------- 
// Forward declarations 
// -------------------- 

// Local includes 

// Declare here #include, #define and typedef 
 
// ----------------- 
// Class declaration 
// ----------------- 
// [!VT]  END 

// [!VT]  HELP 
/** \brief WingDyn_High component class

See the documentation (click the button or refer to vsTaskerDevGuide)
*/  
// [!VT]  END 

// [!VL]  FREQUENCY   30 hz
// [!VL]  INHERIT     BasicDyn
// [!VL]  USES HiAltCtrl, HiSpdCtrl, HiHdgCtrl, HiPitchCtrl, HiRollCtrl 

//====================================================  
class WingDyn_High : public BasicDyn
{  
// [!VT]  CLASS  

/* Declare here your component data and methods 
   Do NOT define Constructors or Destructor!! 
   Use //&& after any data for automatic user-interface: 
   LBL[label] DEF[default] MMX[min-max] HINT[] LIST[] TYPE[] 
   SET[function] INTF[class] UNIT[unit] 
*/ 
 
// Method Section: 
public: 
 
	void autoPilot(float); 
	void manualPilot(float); 
	void FAC(); 
 
	virtual void clampZ(); 
	virtual void computeVelocity();

   virtual double timeTo(WCoord&);        //!< XYZ
   virtual double timeTo2D(WCoord&);      //!< only XY

	virtual void setSpeed(double);          //!< set the target speed with m/s 
	virtual void setHeading(double);        //!< set the target heading, in radian 
	virtual void setHeading(WCoord&);       //!< set the heading target so that to face the given position x,y 
	virtual void setAltitude(double);       //!< set the target altitude in meters 
	virtual void setAltitude(WCoord&);      //!< set the target altitude as to be the one of the given coordinates 
	virtual void setElevation(double);      //!< set the target elevation (slope) in radian 
	virtual void setElevation(WCoord&);     //!< set the target elevation in order to reach the given altitude in coordinates 
	virtual void setRotation(double);       //!< set the target rotation in radian (0 = flat, 180 = upside down) 
 
	virtual void incSpeed(double);          //!< increment the target speed with m/s 
	virtual void incHeading(double);        //!< increment the target heading with radian 
 
   virtual double getBearingTo(WCoord&);   //!< returns the relative bearing to get ther 
   virtual double getSiteTo(WCoord&);      //!< returns the relative elevation to get there 
 
	virtual void incAltitude(double);		 //!< increment the target altitude with meters 
	virtual void incElevation(double);		 //!< increment the target elevation (slope) in radian 
	virtual void incRotation(double);		 //!< increment the target rotation in radian 

	virtual void forceAltitude(double alt);
	virtual void forceHeading(double hdg);
	virtual void forceHeading(WCoord& pos);
	virtual void forceSpeed(double spd);

	virtual void setSpoilers(float);       //!< set here the temporary spoiler increase in m/s

	virtual double getSpeed();				   //!< returns current TAS speed (in m/s) 
	virtual double getHeading();			   //!< returns current heading/azimuth (in radian) 
	virtual double getAltitude();			   //!< returns current altitude (in meters) 
	virtual double getElevation();			//!< returns current elevation (in radian) 
	virtual double getRotation();			   //!< returns current rotation (in radian) 
 
	virtual double getTargetSpeed();		   //!< returns target speed (in m/s) 
	virtual double getTargetHeading();		//!< returns target heading/azimuth (in radian) 
	virtual double getTargetAltitude();		//!< returns target altitude (in meters) 
	virtual double getTargetElevation();	//!< returns target elevation (in radian) 
	virtual double getTargetRotation();		//!< returns target rotation (in radian) 

	virtual int speedOk(float);			    //!< Returns 1 if |current speed - requested speed| < param (m/s)
   virtual int headingOk(float);		       //!< Returns 1 if |current heading - requested heading| < param (deg)
   virtual int altitudeOk(float);		    //!< Returns 1 if |current altitude - requested altitude| < param (m)
 
	float getVerticalSpeed();
	float getGroundSpeed();
	float getTAS();                         //!< return the true air speed (same as getSpeed())
	float getFPA();						       //!< return the flight path angle, in radians

protected: 
 
private: 
 
// Data & Interface Section: 
public: 

	HiAltCtrl   altitude;	 //&& INTF[HiAltCtrl]    DEF[min(0), max(12000), cruise(110), climb_rate(200), descent_rate(300), dt_to_max_rate(1000)]
	HiSpdCtrl   speed;       //&& INTF[HiSpdCtrl]    DEF[min(0), max(140), acc(20), dec(5), dt_to_max_acc(100)]
	HiHdgCtrl   heading;     //&& INTF[HiHdgCtrl]    DEF[min(-180), max(180), max_rate(50), dt_to_max_rate(40)]
	HiPitchCtrl elevation;   //&& INTF[HiPitchCtrl]  DEF[min(-60), max(60), max_rate(20)]
	HiRollCtrl  rotation;    //&& INTF[HiRollCtrl]   DEF[min(-50), max(50), max_rate(10)]

	float V2;  //&& LBL[TakeOff Speed]  DEF[31] UNIT[m/s] 
 
protected: 
	WCoord ps; 
 
private: 
	float d_rate;
	float d_pitch;  
	float d_rotation; 
	float pitch_dec; 
	float pitch_acc; 
	float spoilers;
   WCoord old_pos;
	float wait_pos;

public: 

//! \brief Description 
//!  

protected: 
// [!VT]  END 
  public:  
     WingDyn_High();
     WingDyn_High(int id);

     virtual int  tic();
     virtual void init(int _phase, MData* _mdata);
     virtual void setDefault();
     virtual void setUserCode();
     virtual void setUserData(UsrInterface*);
	  virtual int  react(int, MData*);
     virtual void update(char* =0);
	  virtual void move(); 

  public: // Interface Generated
     void setV2(float);
     void setSpeed(HiSpdCtrl&);
     void setAltitude(HiAltCtrl&);
     void setHeading(HiHdgCtrl&);
     void setElevation(HiPitchCtrl&);
     void setRotation(HiRollCtrl&);

  protected: 

  private: 

}; 

// Declare here all component inline functions 
// Do not forget to write Component|Cpt::method(... 
 
inline void WingDyn_High::setSpeed(double val)     { speed.target = val; } 
inline void WingDyn_High::setHeading(double val)   { heading.target = val; } 
inline void WingDyn_High::setAltitude(double val)  { altitude.target = val; } 
inline void WingDyn_High::setAltitude(WCoord& ps)  { altitude.target = ps.alt; } 
inline void WingDyn_High::setElevation(double val) { elevation.target = val; } 
inline void WingDyn_High::setElevation(WCoord& ps) { elevation.target = pos->elevationTo(ps); } 
inline void WingDyn_High::setRotation(double val)  { rotation.target = val; } 
 
inline void WingDyn_High::incSpeed(double val)     { speed.target += val; } 
inline void WingDyn_High::incHeading(double val)   { heading.target += val; } 
 
inline void WingDyn_High::incAltitude(double val)  { altitude.target += val; } 
inline void WingDyn_High::incElevation(double val) { elevation.target += val; } 
inline void WingDyn_High::incRotation(double val)  { rotation.target += val; } 
 
inline void WingDyn_High::forceAltitude(double alt) { altitude.current = altitude.target = alt; }
inline void WingDyn_High::forceHeading(double hdg)  { heading.current = heading.target = hdg; }
inline void WingDyn_High::forceSpeed(double spd)    { speed.current = speed.target = spd; }

inline void WingDyn_High::setSpoilers(float val) { spoilers = val; }

inline double WingDyn_High::getSpeed()     { return getTAS(); } 
inline double WingDyn_High::getHeading()   { return heading.current; } 
inline double WingDyn_High::getAltitude()  { return altitude.current; } 
inline double WingDyn_High::getElevation() { return elevation.current; } 
inline double WingDyn_High::getRotation()  { return rotation.current; } 
 
inline double WingDyn_High::getTargetSpeed()     { return speed.target; } 
inline double WingDyn_High::getTargetHeading()   { return heading.target; } 
inline double WingDyn_High::getTargetAltitude()  { return altitude.target; } 
inline double WingDyn_High::getTargetElevation() { return elevation.target; } 
inline double WingDyn_High::getTargetRotation()  { return rotation.target; } 

inline void WingDyn_High::setSpeed(HiSpdCtrl& data)       { speed = data; }
inline void WingDyn_High::setAltitude(HiAltCtrl& data)    { altitude = data; }
inline void WingDyn_High::setHeading(HiHdgCtrl& data)     { heading = data; }
inline void WingDyn_High::setElevation(HiPitchCtrl& data) { elevation = data; }
inline void WingDyn_High::setRotation(HiRollCtrl& data)   { rotation = data; }
inline void WingDyn_High::setV2(float data)               { V2 = data; }

inline float WingDyn_High::getVerticalSpeed() { return velocity.z; }
inline float WingDyn_High::getGroundSpeed()   { return sqrt(SQR(velocity.x)+SQR(velocity.y)); }
inline float WingDyn_High::getTAS() { return sqrt(SQR(velocity.x)+SQR(velocity.y)+SQR(velocity.z)); }

#endif 
