
/////////////////////////////////////////////////////////////////////
// ------------------------------------------------------------------
// COPYRIGHT (c) 2004-2016 <VirtualSim>
// ------------------------------------------------------------------
// 
// ------------------------------------------------------------------
// Licensing:
//
/////////////////////////////////////////////////////////////////////


// ------------------
// Include directives
// ------------------
// System Includes
#include <sstream>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <math.h>

// Project Includes

// Local Includes 
#include <basic_dyn.h>
#include <ptf_status.h>
#include <macros.h>
#include <geometry.h>
#include <vt_entities.h>

// --------------------
// Forward declarations
// --------------------

// -------------
// Global data
// -------------

static char* lic = "KJBGAFDSGHITGC";


// ------------------
// Member definitions
// ------------------

#include "array.cpp"



//**********************************
BasicDyn::BasicDyn(int id) : 
   Vt_Component(id), 
   control_mode(AP_Auto),
   altitude_mode(AP_Altitude),
   state(ST_Normal),
   use_gc(NO),
   frozen(NO),
   pos(NULL)
{ 
	memset(&crash, 0, sizeof(CrashData)); 
   hat_deprecated = 0;
}

//**********************************
void BasicDyn::setDefault() 
{ 
   altitudeMode(AP_Altitude);	// deprecated
   controlMode(AP_Auto);		// deprecated
   setState(ST_Normal);			// deprecated
   setDefSpeed(10);				// deprecated
   use_gc = NO;					// deprecated
}

// ************************************************* 
int BasicDyn::react(int _type, MData* _mdata)
{
   return 0;
}

//**********************************
void BasicDyn::startCrash() 
{ 
	crash.hdg_i = getHeading();
	crash.rot_i = 0;
	crash.spd_v = 0;
	crash.rot_v = 0;
	crash.mode  = 1; 
}

//**********************************
void BasicDyn::setCrashed()
{
   if (entity) {
      if (crash.mode != 2) entity->raiseEvent("_Crashed");
      if (entity->hasStatus()) entity->getStatus()->setDamage(_Wreck);
   }
   crash.mode = 2;
   state = ST_Crashed;
}

//**********************************
void BasicDyn::computeVelocity(float yaw, float pitch, float speed)
{
   float gnd_spd = speed*cos(pitch);

   velocity.x = gnd_spd*sin(yaw);
   velocity.y = gnd_spd*cos(yaw);
   velocity.z = speed*sin(pitch);
}

//**********************************
void BasicDyn::computeVelocity()
{
   computeVelocity(getSpeed(), getHeading(), getElevation());
}

//**********************************
void BasicDyn::processVelocity()
{
   float speed = vectorLength(velocity); 
   float pitch = asin(velocity.y / speed);

   setSpeed(speed);
   setPitch(pitch);

   if (cos(pitch)) {
	  float yaw = asin(velocity.x / (cos(pitch)*speed));
	  setYaw(yaw);
   }
}

//**********************************
void BasicDyn::forceVelocity()
{
   float speed = vectorLength(velocity); 
   float pitch = asin(velocity.y / speed);

   forceSpeed(speed);
   setPitch(pitch);

   if (cos(pitch)) {
	  float yaw = asin(velocity.x / (cos(pitch)*speed));
	  setYaw(yaw);
   }
}


//**********************************
WCoord& BasicDyn::getPosIn(float time)
{
   static WCoord next;

   float distance = getSpeed()*time;
   next = getPos()->getNewPosAt(getAzim(), getElev(), distance);
   return next;
}

//**********************************
void BasicDyn::stabilize()
{
   setElevation(0);
   setRotation(0);
   setSpeed(getSpeed());
   setHeading(getHeading());
   setAltitude(getAltitude());
}

//**********************************
int BasicDyn::tic()
{
   return AGAIN;
}

// *************************************************
void BasicDyn::setUserData(UsrInterface* setup)
{
   if (!setup) return;
   // Setup local
   CInterface* intfc = setup->get("BasicDyn");
   if (!intfc) { 
	  fprintf(fout, "Error BasicDyn::setUserData(%s)!\n", setup->getModuleName()); 
	  return; 
   }
   char* data;
   data = intfc->getValue("hat"); if (GOTSTR(data)) hat_deprecated = atof(data); 

   data = intfc->getValue("altitude_mode");
     if (strEqual(data, "AP_Altitude"))  altitudeMode(AP_Altitude);
     if (strEqual(data, "AP_Slope"))     altitudeMode(AP_Slope);
     if (strEqual(data, "AP_Alt_Slope")) altitudeMode(AP_Alt_Slope);

   data = intfc->getValue("state");
     if (strEqual(data, "Normal"))   state = ST_Normal;
     if (strEqual(data, "Grounded")) state = ST_Grounded;
     if (strEqual(data, "Crashed"))  state = ST_Crashed;

   data = intfc->getValue("use_gc");
     if (strEqual(data, "Yes")) use_gc = YES;
     if (strEqual(data, "No"))  use_gc = NO;

   data = intfc->getValue("control_mode");
     if (strEqual(data, "AP_Auto"))     controlMode(AP_Auto);
     if (strEqual(data, "AP_Joystick")) controlMode(AP_Joystick);

   data = intfc->getValue("def_spd"); 
     if (GOTSTR(data)) setDefSpeed(atof(data)); 

   // Put below user code to close settings

}

// *************************************************
void BasicDyn::update(char* ifc)
{ 
   char* data = strGetNew(NAME_SIZE); 
   if (ifc) { 
      Vt_Component::push(TT_NestedIfc, ifc);
   } 
   startUpdate("BasicDyn"); 

   Vt_Component::update("hat", trimFloat(hat_deprecated)); 
   Vt_Component::update("def_spd", trimFloat(def_spd)); 

   CLRSTR(data); 
     if (altitude_mode == AP_Altitude)  strcpy(data, "AP_Altitude");
     if (altitude_mode == AP_Slope)     strcpy(data, "AP_Slope");
     if (altitude_mode == AP_Alt_Slope) strcpy(data, "AP_Alt_Slope");
     if (GOTSTR(data)) Vt_Component::update("altitude_mode", data); 

   CLRSTR(data); 
     if (state == ST_Normal)   strcpy(data, "Normal");
     if (state == ST_Grounded) strcpy(data, "Grounded");
     if (state == ST_Crashed)  strcpy(data, "Crashed");
     if (GOTSTR(data)) Vt_Component::update("state", data); 

   CLRSTR(data); 
     if (use_gc == YES) strcpy(data, "Yes");
     if (use_gc == NO)  strcpy(data, "No");

   CLRSTR(data); 
     if (control_mode == AP_Auto)     strcpy(data, "AP_Auto");
     if (control_mode == AP_Joystick) strcpy(data, "AP_Joystick");
     if (GOTSTR(data)) Vt_Component::update("control_mode", data); 

   endUpdate();
   // Put below update code to wrap-up settings

}

