///////////////////////////////////////////////////////////////////// 
// ------------------------------------------------------------------ 
// COPYRIGHT (c) 2004-2016 <VirtualSim>  
// ------------------------------------------------------------------ 
// 
// ------------------------------------------------------------------ 
// Licensing:  
//  
///////////////////////////////////////////////////////////////////// 

// ------------------  
// Include directives  
// ------------------ 
// System Includes  
#include <sstream> 
#include <stdio.h>  
#include <stdlib.h> 
#include <sys/types.h>  
#include <math.h>  

// Project Includes 

// Local Includes 
#include <wing_dyn_high.h> 
#include <macros.h> 
#include <vt_rtc.h> 
#include <vt_scenarios.h> 

// --------------------  
// Forward declarations  
// --------------------  

// -------------  
// Global data    
// -------------  

// ------------------  
// Member definitions  
// ------------------  


// ************************************************* 
WingDyn_High::WingDyn_High() :
   BasicDyn(0)
{
   init(INIT, NULL);
}

// ************************************************* 
WingDyn_High::WingDyn_High(int id) :
   BasicDyn(id)
{
}

// ************************************************* 
void WingDyn_High::init(int _phase, MData* _mdata)
{
   Vt_SREngine& sre = vt_rtc->srEngine();

   BasicDyn::init(_phase, _mdata);

   // General Component Interface. 
   // Put here code to init the component values. 

   switch (_phase) {

   case INIT: {    // at first start 
      spoilers = 0;
		wait_pos = 0;
   } break;

   case RESET: {   // at each restart 
      if (entity) {
         set(entity->getPos());
         altitude.target = entity->getImport().z;
         altitude.current = altitude.target;
         heading.target = entity->getImport().yaw;
         heading.current = heading.target;
         speed.target = entity->getImport().speed;
         speed.current = speed.target;
			// set velocity vector
			BasicDyn::computeVelocity(heading.get(), elevation.get(), speed.get());
      }
   } break;

   case START: {   // only for start() 
   } break;

   case STOP: {    // only for stop() 
   } break;

   case ENABLE: {
   } break;

   case DISABLE: { // at exit 
   } break;

   case ENTITY_FOCUS: {
   } break;

   case CLEAN: {   // at stop 
   } break;

   case SNAPSHOT: {
   } break;

   case RESTORE: {
   } break;

   case REBUILD: { // after restore 
   } break;
   }
}

// ************************************************* 
// Implement here all your component methods 
// Do not forget to write Component|Cpt::method(... 

void WingDyn_High::move()
{
   double spd = speed.current * deltaT() * cosl(elevation.current);

   if (useGC()) pos->setNewPosAtGC(heading.current, 0, spd);
   else {
      if (heading.current < PI_2) {
         if (heading.current > 0) {
            pos->x += (spd*sinl(heading.current));
            pos->y += (spd*cosl(heading.current));
         }
         else {
            pos->x += (spd*sinl(heading.current));
            pos->y += (spd*cosl(heading.current));
         }
      }
      else {
         if (heading.current > 0) {
            pos->x += (spd*cosl(heading.current - PI_2));
            pos->y -= (spd*sinl(heading.current - PI_2));
         }
         else {
            pos->x -= (spd*cosl(heading.current + PI_2));
            pos->y += (spd*sinl(heading.current + PI_2));
         }
      }
   }
   // spoilers
   altitude.current -= fabs(spoilers) * deltaT();

   if (control_mode == AP_Joystick) {
      // Altitude gain/loss according to speed
      altitude.current += speed.current * deltaT() * sinl(elevation.current);
   }

   pos->z = altitude.current;
}

// ************************************************* 
float WingDyn_High::getFPA()
{
	 float gs = getGroundSpeed();

    if (gs) return atan(getVerticalSpeed()/gs);
    else return PI_2*SGN(getVerticalSpeed());
}

// ************************************************* 
void WingDyn_High::computeVelocity()
{
	// already done in tic()
}

// ************************************************* 
void WingDyn_High::forceHeading(WCoord& pos)
{
   double nhdg;
   if (useGC()) nhdg = entity->getPos()->azimuthToGC(pos);
   else nhdg = entity->getPos()->azimuthTo(pos);
   CHECK_PI(nhdg);
   forceHeading(nhdg);
}

// ************************************************* 
void WingDyn_High::setHeading(WCoord& pos)
{
   double nhdg;
   if (useGC()) nhdg = entity->getPos()->azimuthToGC(pos);
   else nhdg = entity->getPos()->azimuthTo(pos);
   CHECK_PI(nhdg);
   setHeading(nhdg);
}

// ************************************************* 
void WingDyn_High::clampZ()
{
   altitude.current = scenario->terrain.getAlt(pos) + BasicDyn::hat_deprecated;
   altitude.target = altitude.current;
   pos->z = altitude.target;
}

// ************************************************* 
int WingDyn_High::speedOk(float rval)
{
   return (fabs(speed.current - speed.target) <= 2);
}

// ************************************************* 
int WingDyn_High::headingOk(float rval)
{
   return (fabs(heading.current - heading.target) <= DEG2RAD(3));
}

// ************************************************* 
int WingDyn_High::altitudeOk(float rval)
{
   return (fabs(altitude.current - altitude.target) <= rval);
}

// ************************************************* 
void WingDyn_High::autoPilot(float dt)
{
   // speed
   if (speed.target > speed.max) speed.target = speed.max;
   speed.tic(dt);
	
   // altitude
   if (altitude.target > altitude.max) altitude.target = altitude.max;
   else {
      if (altitude.current < altitude.target) // must climb
      {
         if (altitude_mode == AP_Altitude) {
            float tgt_vs = LINEAR_INTERPOL_BOUND(altitude.target - altitude.current, 0.0, altitude.dt_to_max_rate, 0.0, altitude.climb_rate);
            //elevation.target = asinl(altitude.vs/speed.target); 
            elevation.target = LINEAR_INTERPOL_BOUND(tgt_vs, 0.f, altitude.climb_rate, 0.f, DEG2RAD(elevation.max));
            d_pitch = elevation.target - elevation.current;
            if (d_pitch < DEG2RAD(elevation.max_rate)) elevation.current += d_pitch * dt;
            else elevation.current += DEG2RAD(elevation.max_rate) * dt;
            float pitch_factor = LINEAR_INTERPOL_BOUND(elevation.current, 0.f, DEG2RAD(elevation.max), 0.f, 1.f);
            altitude.d_rate = pitch_factor * MIN(altitude.climb_rate, speed.current);
            altitude.current += altitude.d_rate * dt * altitude.inertia;
         }
         else {
            d_pitch = elevation.target - elevation.current;
            if (d_pitch < DEG2RAD(elevation.max_rate)) elevation.current += d_pitch * dt;
            else elevation.current += DEG2RAD(elevation.max_rate) * dt;
            altitude.d_rate = sin(elevation.current) * speed.current;
            altitude.current += altitude.d_rate * dt * altitude.inertia;
            if (fabs(altitude.current - altitude.target) < 30) altitude_mode = AP_Altitude;
         }
      }
      else {  // must descent
         if (altitude_mode == AP_Altitude) {
            float tgt_vs = LINEAR_INTERPOL_BOUND(altitude.current - altitude.target, 0.0, altitude.dt_to_max_rate, 0.0, altitude.descent_rate);
            //elevation.target = -asinl(altitude.vs/speed.target); 
            elevation.target = LINEAR_INTERPOL_BOUND(tgt_vs, 0.f, altitude.descent_rate, 0.f, DEG2RAD(elevation.min));
            d_pitch = elevation.target - elevation.current;
            if (d_pitch > -DEG2RAD(elevation.max_rate)) elevation.current += d_pitch * dt;
            else elevation.current -= DEG2RAD(elevation.max_rate) * dt;
            float pitch_factor = LINEAR_INTERPOL_BOUND(elevation.current, 0.f, DEG2RAD(elevation.min), 0.f, 1.f);
            altitude.d_rate = pitch_factor * MIN(altitude.descent_rate, speed.current);
            altitude.current -= altitude.d_rate * dt * altitude.inertia;
         }
         else {
            d_pitch = elevation.target - elevation.current;
            if (d_pitch > -DEG2RAD(elevation.max_rate)) elevation.current += d_pitch * dt;
            else elevation.current -= DEG2RAD(elevation.max_rate) * dt;
            altitude.d_rate = sin(elevation.current) * speed.current;
            altitude.current += altitude.d_rate * dt * altitude.inertia;
            if (fabs(altitude.current - altitude.target) < 30) altitude_mode = AP_Altitude;
         }
      }
	}
	altitude.inertial(dt);

   // heading
   CHECK_PI(heading.current);

   if (MIN_AGL_SGN(heading.current, heading.target) < 0) {
      d_rate = LINEAR_INTERPOL_BOUND(-MIN_AGL_SGN(heading.current, heading.target), 0.0, DEG2RAD(heading.dt_to_max_rate), 0.0, DEG2RAD(heading.max_rate));
      rotation.target = LINEAR_INTERPOL_BOUND(d_rate, 0.0, DEG2RAD(heading.max_rate), 0.0, DEG2RAD(rotation.max));
      d_rotation = rotation.target - rotation.current;
      if (d_rotation > DEG2RAD(rotation.max_rate)) rotation.current += DEG2RAD(rotation.max_rate * dt);
		else rotation.current += d_rotation * dt;
      if (heading.current != heading.target) {
         heading.current += LINEAR_INTERPOL_BOUND(rotation.current, 0, DEG2RAD(rotation.max), 0, DEG2RAD(heading.max_rate)) * dt;
      }
   }
   else {
      if (MIN_AGL_SGN(heading.current, heading.target) > 0) {
         d_rate = LINEAR_INTERPOL_BOUND(MIN_AGL_SGN(heading.current, heading.target), 0.0, DEG2RAD(heading.dt_to_max_rate), 0.0, DEG2RAD(heading.max_rate));
         rotation.target = LINEAR_INTERPOL_BOUND(d_rate, 0.0, DEG2RAD(heading.max_rate), 0.0, 1.0) * DEG2RAD(rotation.min);
         d_rotation = rotation.target - rotation.current;
         if (d_rotation < -DEG2RAD(rotation.max_rate)) rotation.current -= DEG2RAD(rotation.max_rate * dt);
         else rotation.current += d_rotation * dt;
         if (heading.current != heading.target) {
            heading.current -= LINEAR_INTERPOL_BOUND(-rotation.current, 0, DEG2RAD(-rotation.min), 0, DEG2RAD(heading.max_rate)) * dt;
         }
      }
	}
}

// ************************************************* 
void WingDyn_High::manualPilot(float dt)
{
   if (speed.current < V2) {
      elevation.target = DEG2RAD(-90);
   }

   if (elevation.current <= elevation.target / 2) {
      if (speed.target > speed.max) speed.target = speed.max;
      speed.tic(dt);
   }

   if (elevation.target > elevation.current) {
      d_pitch = elevation.target - elevation.current;
      if (d_pitch < DEG2RAD(elevation.max_rate)) elevation.current += d_pitch * dt;
      else elevation.current += DEG2RAD(elevation.max_rate) * dt;
   }
   else {
      d_pitch = elevation.target - elevation.current;
      if (d_pitch > -DEG2RAD(elevation.max_rate)) elevation.current += d_pitch * dt;
      else elevation.current -= DEG2RAD(elevation.max_rate) * dt;
   }

   float pitch_factor = LINEAR_INTERPOL_BOUND(elevation.current, 0.f, DEG2RAD(elevation.max), 0.f, 1.f);
   altitude.d_rate = pitch_factor * MIN(altitude.climb_rate, speed.current);
   altitude.current += altitude.d_rate * dt;

   if (rotation.target > rotation.current) {
      d_rotation = rotation.target - rotation.current;
      if (d_rotation > DEG2RAD(rotation.max_rate)) rotation.current += DEG2RAD(rotation.max_rate) * dt;
      else rotation.current += d_rotation * dt;
      heading.current += LINEAR_INTERPOL_BOUND(rotation.current, 0, DEG2RAD(rotation.max), 0, DEG2RAD(heading.max_rate)) * dt;
   }
   else {
      d_rotation = rotation.target - rotation.current;
      if (d_rotation < -DEG2RAD(rotation.max_rate)) rotation.current -= DEG2RAD(rotation.max_rate) * dt;
      else rotation.current += d_rotation * dt;
      heading.current -= LINEAR_INTERPOL_BOUND(-rotation.current, 0, DEG2RAD(-rotation.min), 0, DEG2RAD(heading.max_rate)) * dt;
   }
}

// ************************************************* 
int WingDyn_High::react(int _type, MData* _mdata)
{
   return BasicDyn::react(_type, _mdata);
}

// **************************************************      
double WingDyn_High::getBearingTo(WCoord& loc)
{
   double alpha;
   if (useGC()) alpha = pos->azimuthToGC(loc);
   else alpha = pos->azimuthTo(loc);
   return MIN_AGL_SGN(alpha, getAzim());
}

// **************************************************       
double WingDyn_High::getSiteTo(WCoord& loc)
{
   double alpha = pos->elevationTo(loc);
   return MIN_AGL_SGN(alpha, getElev());
}

//**************************************************
double WingDyn_High::timeTo(WCoord& lpos)
{
   if (getSpeed()) return fabs(lpos.distanceTo(*pos) / getSpeed());
   else return HUGE_NUMBER;
}

//**************************************************
double WingDyn_High::timeTo2D(WCoord& lpos)
{
   if (getSpeed()) return fabs(lpos.distanceTo2D(*pos) / getSpeed());
   else return HUGE_NUMBER;
}

// ************************************************* 
int WingDyn_High::tic()
{
   // Implement here the main component runtime function. 
   // Return STANDBY when component is idle or not setup. 

	if (!wait_pos) {
		old_pos = *pos;
	}

   if (control_mode == AP_Auto) autoPilot(deltaT());
   else manualPilot(deltaT());

   move();

   BasicDyn::tic();

	// compute the real velocity vector by substracting the previous t0 position 
	// to the current t1 position, dividing by t1-t0 (elapsed time between the two positions)
	// to get the speed in m/s
	wait_pos += deltaT();
	double lld = old_pos.distanceTo2DGC(*pos);
	if (wait_pos > 0.5 || lld > 3) {
		velocity.x = (pos->x - old_pos.x);
		velocity.y = (pos->y - old_pos.y);
		velocity.z = (pos->z - old_pos.z);
		if (useGC()) {
			double mag = velocity.norm();
			if (mag) {
				double ratio = lld/mag;
				velocity.x *= ratio;
				velocity.y *= ratio;
			}
		}
		velocity /= wait_pos; // t1-t0
		wait_pos = 0;
	}

   return AGAIN;
}

// *************************************************
void WingDyn_High::setDefault()
{
   BasicDyn::setDefault();

   speed.setDefault();				// deprecated
   altitude.setDefault();			// deprecated
   heading.setDefault();			// deprecated
   elevation.setDefault();			// deprecated
   rotation.setDefault();			// deprecated
   setV2((float)31);				// deprecated
}

// *************************************************
void WingDyn_High::setUserData(UsrInterface* setup)
{
   if (!setup) return;
   // Setup parent
   BasicDyn::setUserData(setup);
   // Setup local
   CInterface* intfc = setup->get("WingDyn_High");
   if (!intfc) {
      printf("Error WingDyn_High::setUserData() with %s!\n", setup->getModuleName());
      return;
   }
   char* data = NULL;
   speed.setUserData(intfc->getUsrInterface("speed"));
   altitude.setUserData(intfc->getUsrInterface("altitude"));
   heading.setUserData(intfc->getUsrInterface("heading"));
   elevation.setUserData(intfc->getUsrInterface("elevation"));
   rotation.setUserData(intfc->getUsrInterface("rotation"));
   data = intfc->getValue("V2"); if (GOTSTR(data)) setV2(atof(data));

   // Put below user code to wrap-up settings

}

// *************************************************
void WingDyn_High::update(char* ifc)
{
   // Update parent
   BasicDyn::update();

   char* data = strGetNew(NAME_SIZE);
   if (ifc) {
      Vt_Component::push(TT_NestedIfc, ifc);
   }
   startUpdate("WingDyn_High");

   speed.update("speed");
   altitude.update("altitude");
   heading.update("heading");
   elevation.update("elevation");
   rotation.update("rotation");
   Vt_Component::update("V2", trimFloat(V2));

   endUpdate();
   // Put below update code to wrap-up settings

}

